﻿
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;

namespace opentkprojekt
{
    public partial class Form1 : Form
    {
        bool loaded = false;
        bool animating = false;
        bool rendering = false;

        KeyFraming.KeyFrameArray kfa = new KeyFraming.KeyFrameArray();
        Interpolation.Interpolator animationInterpolator = new Interpolation.CatmullRomInterpolator();

        int animStartTime = 0;

        public Form1()
        {
            InitializeComponent();

            RescaleAnimationTrackbar();
            UpdateAnimationTrackbar();
            InterfaceControlsStatus();

            ChangeInterpolatorName();
        }

        private void ChangeInterpolatorName()
        {
            interpolatorChangeMenu.Text = "Výber interpolácie: " + animationInterpolator.InterpolatorName();
        }

        private void glControl1_Load(object sender, EventArgs e)
        {
            loaded = true;

            GL.ClearColor(Color.SkyBlue);
            SetupViewport();
        }

        private void glControl1_Resize(object sender, EventArgs e)
        {
            if (!loaded)
                return;

            SetupViewport();
            glControl1.Invalidate();
        }

        private void Render()
        {
            if (!loaded) { return; }
            if (rendering) { return; }

            rendering = true;

            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

            if (glControl1.Width == 0 || glControl1.Height == 0) { return; }

            GL.Enable(EnableCap.DepthTest);

            double objTranslX = Helpers.Convertor.StringToDouble(objectTranslX.Text);
            double objTranslY = Helpers.Convertor.StringToDouble(objectTranslY.Text);
            double objTranslZ = Helpers.Convertor.StringToDouble(objectTranslZ.Text);

            Vector3 eye = new Vector3((float)Helpers.Convertor.StringToDouble(cameraX.Text), (float)Helpers.Convertor.StringToDouble(cameraY.Text), (float)Helpers.Convertor.StringToDouble(cameraZ.Text));
            Vector3 camTarget = new Vector3((float)Helpers.Convertor.StringToDouble(cameraPointX.Text), (float)Helpers.Convertor.StringToDouble(cameraPointY.Text), (float)Helpers.Convertor.StringToDouble(cameraPointZ.Text));
            Matrix4 modelview = Matrix4.LookAt(eye, camTarget, Vector3.UnitY);

            GL.MatrixMode(MatrixMode.Modelview);
            GL.LoadMatrix(ref modelview);

            Vector3d rotationVector = new Vector3d(Helpers.Convertor.StringToDouble(objectRotX.Text), Helpers.Convertor.StringToDouble(objectRotY.Text), Helpers.Convertor.StringToDouble(objectRotZ.Text));
            rotationVector.Normalize();
            double rotAngle = Helpers.Convertor.DegToRad(Helpers.Convertor.StringToDouble(objectRotW.Text));
            Quaterniond rotation = new Quaterniond(rotationVector * Math.Sin(rotAngle / 2.0), Math.Cos(rotAngle / 2.0));

            GL.Translate(objTranslX, objTranslY, objTranslZ);

            GL.Scale(Helpers.Convertor.StringToDouble(objectScaleX.Text), Helpers.Convertor.StringToDouble(objectScaleY.Text), Helpers.Convertor.StringToDouble(objectScaleZ.Text));

            if (rotation.Length > 0)
            {
                Matrix4d rot = Matrix4d.CreateFromQuaternion(rotation);
                GL.MultMatrix(ref rot);
            }
                        
            GL.Begin(PrimitiveType.Quads);
            GL.Color3(Color.Purple);
            GL.Vertex3(1, 1, 1);
            GL.Vertex3(-1, 1, 1);
            GL.Vertex3(-1, -1, 1);
            GL.Vertex3(1, -1, 1);
            GL.Color3(Color.Red);
            GL.Vertex3(1, 1, -1);
            GL.Vertex3(-1, 1, -1);
            GL.Vertex3(-1, -1, -1);
            GL.Vertex3(1, -1, -1);
            GL.Color3(Color.Orange);
            GL.Vertex3(1, 1, 1);
            GL.Vertex3(1, 1, -1);
            GL.Vertex3(1, -1, -1);
            GL.Vertex3(1, -1, 1);
            GL.Color3(Color.Green);
            GL.Vertex3(-1, 1, 1);
            GL.Vertex3(-1, 1, -1);
            GL.Vertex3(-1, -1, -1);
            GL.Vertex3(-1, -1, 1);
            GL.Color3(Color.Blue);
            GL.Vertex3(1, 1, 1);
            GL.Vertex3(-1, 1, 1);
            GL.Vertex3(-1, 1, -1);
            GL.Vertex3(1, 1, -1);
            GL.Color3(Color.Tomato);
            GL.Vertex3(1, -1, 1);
            GL.Vertex3(-1, -1, 1);
            GL.Vertex3(-1, -1, -1);
            GL.Vertex3(1, -1, -1);
            GL.End();

            glControl1.SwapBuffers();

            rendering = false;
        }

        private void glControl1_Paint(object sender, PaintEventArgs e)
        {
            Render();
        }

        private void SetupViewport()
        {
            int w = glControl1.Width;
            int h = glControl1.Height;
            GL.Viewport(0, 0, w, h); // Use all of the glControl painting area
            GL.MatrixMode(MatrixMode.Projection);
            if (h != 0 && w != 0) 
            { 
                Matrix4 projection = Matrix4.CreatePerspectiveFieldOfView((float)Math.PI / 4, w / (float)h, 1.0f, 64.0f);
                GL.LoadMatrix(ref projection);
            }
        }

        private void DoRender()
        {
            if (!animating)
            {
                glControl1.Invalidate();
            }
        }

        private void InterfaceControlsStatus()
        {
            cameraX.Enabled = !playAnimationTimer.Enabled;
            cameraY.Enabled = !playAnimationTimer.Enabled;
            cameraZ.Enabled = !playAnimationTimer.Enabled;

            cameraPointX.Enabled = !playAnimationTimer.Enabled && !cameraFollowsObject.Checked;
            cameraPointY.Enabled = !playAnimationTimer.Enabled && !cameraFollowsObject.Checked;
            cameraPointZ.Enabled = !playAnimationTimer.Enabled && !cameraFollowsObject.Checked;

            cameraFollowsObject.Enabled = !playAnimationTimer.Enabled;

            objectTranslX.Enabled = !playAnimationTimer.Enabled;
            objectTranslY.Enabled = !playAnimationTimer.Enabled;
            objectTranslZ.Enabled = !playAnimationTimer.Enabled;

            objectScaleX.Enabled = !playAnimationTimer.Enabled;
            objectScaleY.Enabled = !playAnimationTimer.Enabled;
            objectScaleZ.Enabled = !playAnimationTimer.Enabled;

            objectRotX.Enabled = !playAnimationTimer.Enabled;
            objectRotY.Enabled = !playAnimationTimer.Enabled;
            objectRotZ.Enabled = !playAnimationTimer.Enabled;
            objectRotW.Enabled = !playAnimationTimer.Enabled;

            saveKeyframe.Enabled = !playAnimationTimer.Enabled;

            addFramesButton.Enabled = !playAnimationTimer.Enabled;
            addFramesTextBox.Enabled = !playAnimationTimer.Enabled;

            framerateInput.Enabled = !playAnimationTimer.Enabled;
            framerateChangeButton.Enabled = !playAnimationTimer.Enabled;

            newAnimation.Enabled = !playAnimationTimer.Enabled;

            deleteKeyFrameButton.Enabled = !playAnimationTimer.Enabled && kfa.IsKeyframeInTime(animationTrackbar.Value);

            animationTrackbar.Enabled = !playAnimationTimer.Enabled;
        }

        private void cameraX_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void cameraY_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void cameraZ_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void objectTranslX_TextChanged(object sender, EventArgs e)
        {
            if (cameraFollowsObject.Checked)
            {
                cameraPointX.Text = objectTranslX.Text;
                cameraPointY.Text = objectTranslY.Text;
                cameraPointZ.Text = objectTranslZ.Text;
            }
            DoRender();
        }

        private void objectTranslY_TextChanged(object sender, EventArgs e)
        {
            if (cameraFollowsObject.Checked)
            {
                cameraPointX.Text = objectTranslX.Text;
                cameraPointY.Text = objectTranslY.Text;
                cameraPointZ.Text = objectTranslZ.Text;
            }
            DoRender();
        }

        private void objectTranslZ_TextChanged(object sender, EventArgs e)
        {
            if (cameraFollowsObject.Checked)
            {
                cameraPointX.Text = objectTranslX.Text;
                cameraPointY.Text = objectTranslY.Text;
                cameraPointZ.Text = objectTranslZ.Text;
            }
            DoRender();
        }

        private void cameraPointX_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void cameraPointY_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void cameraPointZ_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void cameraFollowsObject_CheckedChanged(object sender, EventArgs e)
        {
            if (cameraFollowsObject.Checked)
            {
                cameraPointX.Text = objectTranslX.Text;
                cameraPointY.Text = objectTranslY.Text;
                cameraPointZ.Text = objectTranslZ.Text;
            }
            InterfaceControlsStatus();
            DoRender();
        }

        private void objectScaleX_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void objectScaleY_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void objectScaleZ_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void objectRotX_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void objectRotY_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void objectRotZ_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void objectRotW_TextChanged(object sender, EventArgs e)
        {
            DoRender();
        }

        private void RescaleAnimationTrackbar()
        {
            animationTrackbar.Maximum = Math.Max(kfa.GetAnimationLength() - 1, 0);
        }

        private void animationTrackbar_Scroll(object sender, EventArgs e)
        {
            UpdateAnimationTrackbar();
            InterfaceControlsStatus();
        }

        private void UpdateAnimationTrackbar()
        {
            animating = true;
            var animationProperties = animationInterpolator.Interpolate(animationTrackbar.Value, ref kfa);
            setAnimationProperties(animationProperties);
            if (kfa.IsKeyframeInTime(animationTrackbar.Value))
            {
                animationTrackbar.BackColor = System.Drawing.Color.Red;
            }
            else
            {
                animationTrackbar.BackColor = System.Drawing.Color.Silver;
            }
            keyframeLabel.Text = "Rámec: " + Convert.ToSingle(animationTrackbar.Value) + " (" + Convert.ToString(kfa.GetFPS()) + " FPS)";
            animating = false;
            glControl1.Invalidate();
        }

        private void setAnimationProperties(KeyValuePair<KeyFraming.ObjectKeyFrame, KeyFraming.CameraKeyFrame> animationProperties)
        {
            cameraFollowsObject.Checked = animationProperties.Value.GetFollowTarget();
            cameraX.Text = Convert.ToString(animationProperties.Value.GetPosition().X);
            cameraY.Text = Convert.ToString(animationProperties.Value.GetPosition().Y);
            cameraZ.Text = Convert.ToString(animationProperties.Value.GetPosition().Z);
            cameraPointX.Text = Convert.ToString(animationProperties.Value.GetTarget().X);
            cameraPointY.Text = Convert.ToString(animationProperties.Value.GetTarget().Y);
            cameraPointZ.Text = Convert.ToString(animationProperties.Value.GetTarget().Z);
            objectTranslX.Text = Convert.ToString(animationProperties.Key.GetTranslation().X);
            objectTranslY.Text = Convert.ToString(animationProperties.Key.GetTranslation().Y);
            objectTranslZ.Text = Convert.ToString(animationProperties.Key.GetTranslation().Z);
            objectScaleX.Text = Convert.ToString(animationProperties.Key.GetScale().X);
            objectScaleY.Text = Convert.ToString(animationProperties.Key.GetScale().Y);
            objectScaleZ.Text = Convert.ToString(animationProperties.Key.GetScale().Z);
            double angle = Math.Acos(animationProperties.Key.GetRotationQuaternion().W);
            Vector3d rotXYZ = animationProperties.Key.GetRotationQuaternion().Xyz;
            if (rotXYZ.Length != 0)
            {
                rotXYZ.Normalize();
            }
            objectRotX.Text = Convert.ToString(rotXYZ.X);
            objectRotY.Text = Convert.ToString(rotXYZ.Y);
            objectRotZ.Text = Convert.ToString(rotXYZ.Z);
            objectRotW.Text = Convert.ToString(Helpers.Convertor.RadToDeg(angle * 2));
        }

        private void saveKeyframe_Click(object sender, EventArgs e)
        {
            KeyFraming.CameraKeyFrame ckf = new KeyFraming.CameraKeyFrame();
            KeyFraming.ObjectKeyFrame okf = new KeyFraming.ObjectKeyFrame();

            Vector3d rotationVector = new Vector3d(Helpers.Convertor.StringToDouble(objectRotX.Text), Helpers.Convertor.StringToDouble(objectRotY.Text), Helpers.Convertor.StringToDouble(objectRotZ.Text));
            rotationVector.Normalize();
            double rotAngle = Helpers.Convertor.DegToRad(Helpers.Convertor.StringToDouble(objectRotW.Text));
            Quaterniond rotation = new Quaterniond(rotationVector * Math.Sin(rotAngle / 2.0), Math.Cos(rotAngle / 2.0));

            if (rotation.Length > 0)
            {
                okf.SetRotation(rotation);
            }
            else
            {
                okf.SetRotation(new Quaterniond(0, 0, 0, 1));
            }

            double objTranslX = Helpers.Convertor.StringToDouble(objectTranslX.Text);
            double objTranslY = Helpers.Convertor.StringToDouble(objectTranslY.Text);
            double objTranslZ = Helpers.Convertor.StringToDouble(objectTranslZ.Text);
            okf.SetTranslation(new Vector3d(objTranslX, objTranslY, objTranslZ));
            okf.SetScale(new Vector3d(Helpers.Convertor.StringToDouble(objectScaleX.Text), Helpers.Convertor.StringToDouble(objectScaleY.Text), Helpers.Convertor.StringToDouble(objectScaleZ.Text)));

            ckf.SetFollowTarget(cameraFollowsObject.Checked);
            ckf.SetPosition(new Vector3d(Helpers.Convertor.StringToDouble(cameraX.Text), Helpers.Convertor.StringToDouble(cameraY.Text), Helpers.Convertor.StringToDouble(cameraZ.Text)));
            if (cameraFollowsObject.Checked)
            {
                ckf.SetTarget(new Vector3d(objTranslX, objTranslY, objTranslZ));
            }
            else
            {
                ckf.SetTarget(new Vector3d(Helpers.Convertor.StringToDouble(cameraPointX.Text), Helpers.Convertor.StringToDouble(cameraPointY.Text), Helpers.Convertor.StringToDouble(cameraPointZ.Text)));
            }

            kfa.SaveKeyframe(animationTrackbar.Value, okf, ckf);
            UpdateAnimationTrackbar();
            InterfaceControlsStatus();
        }

        private void menuStrip1_ItemClicked(object sender, ToolStripItemClickedEventArgs e)
        {

        }

        private void interpolateCatmullRom_Click(object sender, EventArgs e)
        {
            animationInterpolator = new Interpolation.CatmullRomInterpolator();
            interpolateCatmullRom.Checked = true;
            interpolateLinear.Checked = false;
            interpolateNN.Checked = false;
            UpdateAnimationTrackbar();
            ChangeInterpolatorName();
        }

        private void interpolateLinear_Click(object sender, EventArgs e)
        {
            animationInterpolator = new Interpolation.LinearInterpolator();
            interpolateCatmullRom.Checked = false;
            interpolateLinear.Checked = true;
            interpolateNN.Checked = false;
            UpdateAnimationTrackbar();
            ChangeInterpolatorName();
        }

        private void interpolateNN_Click(object sender, EventArgs e)
        {
            animationInterpolator = new Interpolation.NearestNeighborInterpolator();
            interpolateCatmullRom.Checked = false;
            interpolateLinear.Checked = false;
            interpolateNN.Checked = true;
            UpdateAnimationTrackbar();
            ChangeInterpolatorName();
        }

        private void addFramesButton_Click(object sender, EventArgs e)
        {
            int time = (int)Helpers.Convertor.StringToDouble(addFramesTextBox.Text);
            if (time > 0)
            {
                kfa.AddAnimationLength(time);
                RescaleAnimationTrackbar();
                UpdateAnimationTrackbar();
            }
        }

        private void playAnimationTimer_Tick(object sender, EventArgs e)
        {
            if (animationTrackbar.Value < animationTrackbar.Maximum)
            {
                animationTrackbar.Value = animationTrackbar.Value + 1;
            }
            else
            {
                animationTrackbar.Value = animationTrackbar.Minimum;
            }
            UpdateAnimationTrackbar();
        }

        private void startStopButton_Click(object sender, EventArgs e)
        {
            int timerInterval = kfa.GetTimerTickInterval();
            playAnimationTimer.Interval = timerInterval;
            if (playAnimationTimer.Enabled)
            {
                startStopButton.Text = "Spustiť";
                animationTrackbar.Value = animStartTime;
                UpdateAnimationTrackbar();
            }
            else
            {
                startStopButton.Text = "Zastaviť";
                animStartTime = animationTrackbar.Value;
            }
            playAnimationTimer.Enabled = !playAnimationTimer.Enabled;
            InterfaceControlsStatus();
        }

        private void framerateChangeButton_Click(object sender, EventArgs e)
        {
            if (!playAnimationTimer.Enabled)
            {
                kfa.SetFPS((int)framerateInput.Value);
                RescaleAnimationTrackbar();
                UpdateAnimationTrackbar();
            }
        }

        private void newAnimation_Click(object sender, EventArgs e)
        {
            var answer = MessageBox.Show("Naozaj chcete vytvoriť novú animáciu? Stará animácia sa stratí.", "Nová animácia", MessageBoxButtons.YesNo);
            if (answer == DialogResult.Yes)
            {
                kfa = new KeyFraming.KeyFrameArray();
                animationTrackbar.Value = 0;
                RescaleAnimationTrackbar();
                UpdateAnimationTrackbar();
            }
        }

        private void deleteKeyFrameButton_Click(object sender, EventArgs e)
        {
            kfa.DeleteKeyframe(animationTrackbar.Value);
            UpdateAnimationTrackbar();
            InterfaceControlsStatus();
        }

    }
}
