﻿using System;
using System.Collections.Generic;
using System.Text;

using OpenTK;

namespace opentkprojekt.Interpolation
{
    class LinearInterpolator : Interpolator
    {
        public KeyValuePair<KeyFraming.ObjectKeyFrame, KeyFraming.CameraKeyFrame> Interpolate(int time, ref KeyFraming.KeyFrameArray kfa)
        {
            if (kfa.IsKeyframeInTime(time))
            {
                return kfa.GetKeyframeByTime(time);
            }
            
            int first = 0;
            int second = 0;
            kfa.FindKeyframesIndexes(time, out first, out second);
            int timeFirst = kfa.GetKeyframeTimeByIndex(first);
            int timeSecond = kfa.GetKeyframeTimeByIndex(second);

            double dtime = (double)(time - timeFirst) / (double)(timeSecond - timeFirst);
            if (Double.IsNaN(dtime)) { dtime = 0; }

            KeyFraming.CameraKeyFrame ckf = new KeyFraming.CameraKeyFrame();
            KeyFraming.ObjectKeyFrame okf = new KeyFraming.ObjectKeyFrame();

            var kf1 = kfa.GetKeyframeByIndex(first);
            var kf2 = kfa.GetKeyframeByIndex(second);

            okf.SetTranslation(LinearInterpolation(dtime, kf1.Key.GetTranslation(), kf2.Key.GetTranslation()));

            okf.SetScale(LinearInterpolation(dtime, kf1.Key.GetScale(), kf2.Key.GetScale()));

            var qrot1 = kf1.Key.GetRotationQuaternion();
            var qrot2 = kf2.Key.GetRotationQuaternion();
            okf.SetRotation(SphericalLinearInterpolation.Interpolate(qrot1, qrot2, dtime));

            ckf.SetFollowTarget(false);//kf2.Value.GetFollowTarget());

            ckf.SetPosition(LinearInterpolation(dtime, kf1.Value.GetPosition(), kf2.Value.GetPosition()));

            var target1 = kf1.Value.GetFollowTarget() ? kf1.Key.GetTranslation() : kf1.Value.GetTarget();
            var target2 = kf2.Value.GetFollowTarget() ? kf2.Key.GetTranslation() : kf2.Value.GetTarget();
            ckf.SetTarget(LinearInterpolation(dtime, target1, target2));

            return new KeyValuePair<KeyFraming.ObjectKeyFrame, KeyFraming.CameraKeyFrame>(okf, ckf);
        }

        public OpenTK.Vector3d LinearInterpolation(double time, OpenTK.Vector3d from, OpenTK.Vector3d to)
        {
            return (1 - time) * from + time * to;
        }


        public string InterpolatorName()
        {
            return "Lineárna";
        }
    }
}
