﻿using System;
using System.Collections.Generic;
using System.Text;

using OpenTK;

namespace opentkprojekt.KeyFraming
{
    class ObjectKeyFrame
    {
        protected Vector3d translation = new Vector3d();
        protected Vector3d scale = new Vector3d();
        protected Quaterniond rotation = new Quaterniond();

        public ObjectKeyFrame()
        {
            scale.X = 1.0;
            scale.Y = 1.0;
            scale.Z = 1.0;
            Vector3d pointing = new Vector3d(1, 0, 0);
            pointing.Normalize();
            rotation = new Quaterniond(pointing * Math.Sin(0), Math.Cos(0));
            rotation.Normalize();
        }

        public ObjectKeyFrame(double x, double y, double z, double sX, double sY, double sZ, double rotX, double rotY, double rotZ, double angle)
        {
            translation.X = x;
            translation.Y = y;
            translation.Z = z;
            scale.X = sX;
            scale.Y = sY;
            scale.Z = sZ;
            Vector3d pointing = new Vector3d(rotX, rotY, rotZ);
            pointing.Normalize();
            rotation = new Quaterniond(pointing * Math.Sin(angle / 2.0), Math.Cos(angle / 2.0));
            rotation.Normalize();
        }

        public ObjectKeyFrame(double x, double y, double z, double sX, double sY, double sZ, Matrix3d rotationMatrix)
        {
            translation.X = x;
            translation.Y = y;
            translation.Z = z;
            scale.X = sX;
            scale.Y = sY;
            scale.Z = sZ;
            rotation = Quaterniond.FromMatrix(rotationMatrix);
            rotation.Normalize();
        }

        public void SetTranslation(Vector3d translation)
        {
            this.translation = translation;
        }

        public Vector3d GetTranslation()
        {
            return new Vector3d(translation);
        }

        public void SetScale(Vector3d scale)
        {
            this.scale = scale;
        }

        public Vector3d GetScale()
        {
            return new Vector3d(scale);
        }

        public void SetRotation(Matrix3d rotationMatrix)
        {
            rotation = Quaterniond.FromMatrix(rotationMatrix);
            rotation.Normalize();
        }

        public void SetRotation(Quaterniond rotQuat)
        {
            rotation = rotQuat;
            rotation.Normalize();
        }

        public Quaterniond GetRotationQuaternion()
        {
            return new Quaterniond(rotation.Xyz, rotation.W);
        }

        public Vector3d GetRotationXYZ()
        {
            Vector3d output = new Vector3d();

            double test = rotation.X * rotation.Y + rotation.Z * rotation.W;

            if (test > 0.499)
            {
                output.Y = 2 * Math.Atan2(rotation.X, rotation.W);
                output.Z = Math.PI / 2.0f;
                output.X = 0.0f;
            }
            else if (test < -0.499)
            {
                output.Y = -2 * Math.Atan2(rotation.X, rotation.W);
                output.Z = -Math.PI / 2.0f;
                output.X = 0.0f;
            }
            else
            {
                double sqx = rotation.X * rotation.X;
                double sqy = rotation.Y * rotation.Y;
                double sqz = rotation.Z * rotation.Z;

                output.Y = Math.Atan2(2 * rotation.Y * rotation.W - 2 * rotation.X * rotation.Z, 1 - 2 * sqy - 2 * sqz);
                output.Z = Math.Asin(2 * test);
                output.X = Math.Atan2(2 * rotation.X * rotation.W - 2 * rotation.Y * rotation.Z, 1 - 2 * sqx - 2 * sqz);
            }

            return output;
        }
    }
}
