﻿using System;
using System.Collections.Generic;
using System.Text;

namespace opentkprojekt.KeyFraming
{   
    class KeyFrameArray
    {
        protected int fps = 24;
        protected int length = 25;
        protected SortedList<int, KeyValuePair<ObjectKeyFrame, CameraKeyFrame>> keyframes = new SortedList<int, KeyValuePair<ObjectKeyFrame, CameraKeyFrame>>();

        public KeyFrameArray()
        {
            fps = 24;
            length = 25;
            InsertDefaultKeyFrame(0);
            InsertDefaultKeyFrame(24);
        }

        protected void InsertDefaultKeyFrame(int time)
        {
            if (time > length)
            {
                length = time + 1;
            }
            if (time >= 0)
            {
                KeyValuePair<ObjectKeyFrame, CameraKeyFrame> keyframe = new KeyValuePair<ObjectKeyFrame, CameraKeyFrame>(new ObjectKeyFrame(), new CameraKeyFrame());
                if (keyframes.ContainsKey(time))
                {
                    keyframes.Remove(time);
                }
                keyframes.Add(time, keyframe);
            }
            else
            {
                throw new Exception("Negative time index is not allowed.");
            }
        }

        public void AddAnimationLength(int addLength)
        {
            if (addLength > 0)
            {
                length += addLength;
                int time = length - 1;
                int last = Math.Max(GetKeyframeIndexesCound() - 1, 0);
                var lastKeyframe = GetKeyframeByIndex(last);
                SaveKeyframe(time, lastKeyframe.Key, lastKeyframe.Value);
            }
        }

        public void SaveKeyframe(int time, ObjectKeyFrame okf, CameraKeyFrame ckf)
        {
            if (time > length)
            {
                length = time + 1;
            }
            if (time >= 0)
            {
                KeyValuePair<ObjectKeyFrame, CameraKeyFrame> keyframe = new KeyValuePair<ObjectKeyFrame, CameraKeyFrame>(okf, ckf);
                if (keyframes.ContainsKey(time))
                {
                    keyframes.Remove(time);
                }
                keyframes.Add(time, keyframe);
            }
            else
            {
                throw new Exception("Negative time index is not allowed.");
            }
        }

        public void DeleteKeyframe(int time)
        {
            if (keyframes.ContainsKey(time) && time > 0 && time < length - 1)
            {
                keyframes.Remove(time);
            }
        }

        public int GetAnimationLength()
        {
            return length;
        }

        public bool IsKeyframeInTime(int time)
        {
            return keyframes.ContainsKey(time);
        }

        public KeyValuePair<ObjectKeyFrame, CameraKeyFrame> GetKeyframeByTime(int time)
        {
            if (keyframes.ContainsKey(time))
            {
                return keyframes[time];
            }
            throw new Exception("No keyframe for this time.");
        }

        public KeyValuePair<ObjectKeyFrame, CameraKeyFrame> GetKeyframeByIndex(int index)
        {
            if (index >= 0 && index < keyframes.Keys.Count)
            {
                int time = keyframes.Keys[index];
                return keyframes[time];
            }
            throw new Exception("Index out of bounds.");
        }

        public int GetKeyframeTimeByIndex(int index)
        {
            if (index >= 0 && index < keyframes.Keys.Count)
            {
                return keyframes.Keys[index];
            }
            throw new Exception("Index out of bounds.");
        }

        public int GetKeyframeIndexesCound()
        {
            return keyframes.Keys.Count;
        }

        public void FindKeyframesIndexes(int time, out int first, out int second)
        {
            if (time < 0) 
            {
                first = 0;
                second = 0;
                return;
            }
            if (time >= length)
            {
                first = keyframes.Keys.Count - 1;
                second = first;
                return;
            }
            first = 0;
            second = keyframes.Keys.Count - 1;

            while (second - first != 1)
            {
                int current = (second - first) / 2 + first;
                int indexTime = keyframes.Keys[current];
                if (indexTime == time)
                {
                    first = current;
                    second = current;
                    return;
                }
                else if (indexTime > time)
                {
                    second = current;
                }
                else
                {
                    first = current;
                }
            }
        }

        public int GetFPS()
        {
            return fps;
        }

        public int GetTimerTickInterval()
        {
            return 1000 / GetFPS();
        }

        public void SetFPS(int newFPS)
        {
            if (newFPS == fps) { return;  }
            if (newFPS >= 24 && newFPS <= 60)
            {
                int currentTime = length - 1;
                double currentTimeRatio = (double)currentTime / (double)fps;
                int newTime = (int)(currentTimeRatio * newFPS);
                SortedList<int, KeyValuePair<ObjectKeyFrame, CameraKeyFrame>> newKeyframes = new SortedList<int, KeyValuePair<ObjectKeyFrame, CameraKeyFrame>>();
                SortedList<int, List<int>> newFramesMapping = new SortedList<int, List<int>>();
                foreach (var keyframe in keyframes)
                {
                    var addTo = 0;
                    if (keyframe.Key == 0)
                    {
                        addTo = 0;
                    }
                    else if (keyframe.Key == currentTime)
                    {
                        addTo = newTime;
                    }
                    else
                    {
                        double timeRatio = (double)keyframe.Key / (double)currentTime;
                        int kfNewTime = (int)((1 - timeRatio) * 1 + timeRatio * newTime);
                        addTo = kfNewTime;
                    }
                    if (!newFramesMapping.ContainsKey(addTo))
                    {
                        newFramesMapping.Add(addTo, new List<int>());
                    }
                    newFramesMapping[addTo].Add(keyframe.Key);
                }

                foreach (var kflist in newFramesMapping)
                {
                    newKeyframes.Add(kflist.Key, Approximate(kflist.Value));
                }
                keyframes = newKeyframes;
                fps = newFPS;
                length = newTime + 1;
            }
        }

        private KeyValuePair<ObjectKeyFrame, CameraKeyFrame> Approximate(List<int> times)
        {
            if (times.Count == 1)
            {
                return keyframes[times[0]];
            }
            else
            {
                List<KeyValuePair<ObjectKeyFrame, CameraKeyFrame>> localKeyframes = new List<KeyValuePair<ObjectKeyFrame, CameraKeyFrame>>();
                foreach (var time in times)
                {
                    localKeyframes.Add(keyframes[time]);
                }

                Interpolation.LinearInterpolator lip = new Interpolation.LinearInterpolator();
                while (localKeyframes.Count > 1)
                {
                    List<KeyValuePair<ObjectKeyFrame, CameraKeyFrame>> newLocalKeyframes = new List<KeyValuePair<ObjectKeyFrame, CameraKeyFrame>>();
                    for (int i = 1; i < localKeyframes.Count; i++)
                    {
                        KeyValuePair<ObjectKeyFrame, CameraKeyFrame> first = localKeyframes[i-1];
                        KeyValuePair<ObjectKeyFrame, CameraKeyFrame> second = localKeyframes[i];

                        ObjectKeyFrame okf = new ObjectKeyFrame();
                        CameraKeyFrame ckf = new CameraKeyFrame();

                        okf.SetTranslation(lip.LinearInterpolation(0.5, first.Key.GetTranslation(), second.Key.GetTranslation()));

                        okf.SetScale(lip.LinearInterpolation(0.5, first.Key.GetScale(), second.Key.GetScale()));

                        var qrot1 = first.Key.GetRotationQuaternion();
                        var qrot2 = second.Key.GetRotationQuaternion();
                        okf.SetRotation(SphericalLinearInterpolation.Interpolate(qrot1, qrot2, 0.5));

                        ckf.SetFollowTarget(second.Value.GetFollowTarget());

                        ckf.SetPosition(lip.LinearInterpolation(0.5, first.Value.GetPosition(), second.Value.GetPosition()));

                        var target1 = first.Value.GetFollowTarget() ? first.Key.GetTranslation() : first.Value.GetTarget();
                        var target2 = second.Value.GetFollowTarget() ? second.Key.GetTranslation() : second.Value.GetTarget();
                        ckf.SetTarget(lip.LinearInterpolation(0.5, target1, target2));
                        newLocalKeyframes.Add(new KeyValuePair<ObjectKeyFrame, CameraKeyFrame>(okf, ckf));
                    }
                    localKeyframes = newLocalKeyframes;
                }

                return localKeyframes[0];
            }
        }

    }
}
